fs = require 'fs'
assert = require 'assert'
_ = require 'underscore'
{assertCalledOnce} = require './helpers'

# Constructor for action objects
@action = action = (inPaths, outPaths, perform) ->
  assert inPaths instanceof Array
  assert _.every inPaths, (x) -> typeof x is 'string'
  assert outPaths instanceof Array
  assert _.every outPaths, (x) -> typeof x is 'string'
  assert typeof perform is 'function'
  {inPaths, outPaths, perform}

# Build procedure
@build = (actions...) ->
  t0 = Date.now()

  actions = _.flatten actions

  # If the last argument is a function, use it as a callback
  if actions.length and typeof actions[actions.length - 1] is 'function'
    callback = assertCalledOnce actions.splice(-1)[0]

  # Do some validations
  for x in actions
    assert typeof x is 'object'
    assert x.inPaths instanceof Array
    assert _.every x.inPaths, (x) -> typeof x is 'string'
    assert x.outPaths instanceof Array
    assert _.every x.outPaths, (x) -> typeof x is 'string'
    assert typeof x.perform is 'function'

  # Collect information about files mentioned as inPaths or outPaths
  fileInfos = {} # indexed by path
  for x in actions
    for paths in [x.inPaths, x.outPaths]
      for path in paths
        fileInfos[path] ?= inActions: [], outActions: []
    for path in x.inPaths
      fileInfos[path].outActions.push x
    for path in x.outPaths
      fileInfos[path].inActions.push x

  # Create a graph whose nodes are the action objects.
  # (x, y) is an edge iff the intersection between x.outPaths and y.inPaths is
  # non-empty.
  for x in actions
    x.neighbours = _(x.outPaths).chain()
      .map((path) -> fileInfos[path].outActions)
      .flatten().uniq().value()
    assert x not in x.neighbours, 'Dependency graph contains a loop'

  # Topological sorting
  for x in actions then x.inDegree = 0
  for x in actions then for y in x.neighbours then y.inDegree++
  oldActions = actions # unsorted list
  actions = [] # sorted list
  done = false
  while not done
    done = true
    for x in oldActions when x not in actions and x.inDegree is 0
      done = false
      actions.push x
      for y in x.neighbours then y.inDegree--

  # Cache statSync() calls
  getMTime = _.memoize (path) ->
    try fs.statSync(path).mtime.getTime() catch e then 0

  # Mark dirty actions
  for a in actions when not a.dirty
    latestIn = Math.max (a.inPaths.map getMTime)...
    earliestOut = Math.min (a.outPaths.map getMTime)...
    if earliestOut < latestIn
      a.dirty = true
      # Extend dirtiness to the transitive closure of a:
      queue = [a]
      while queue.length
        x = queue.shift()
        for path in x.outPaths
          for y in fileInfos[path].outActions
            if not y.dirty
              y.dirty = true
              queue.push y

  # Perform actions
  dirtyActions = for x in actions when x.dirty then x
  for x in dirtyActions then for y in x.neighbours then y.inDegree++
  actionsBeingPerformed = []
  do f = ->
    done = true
    dirtyActions.forEach (x) ->
      if not x.inDegree
        done = false
        dirtyActions = _(dirtyActions).without x
        actionsBeingPerformed.push x
        log = (message) -> console.info "#{Date.now() - t0} [#{x.outPaths.join ' '}] #{message}"
        log 'start'
        x.perform
          inPaths: x.inPaths
          outPaths: x.outPaths
          log: log
          callback: assertCalledOnce ->
            log 'done'
            actionsBeingPerformed = _(actionsBeingPerformed).without x
            for y in x.neighbours then y.inDegree--
            f?()
    if done and not actionsBeingPerformed.length
      f = null
      callback?()

  return
